@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -26,12 +26,14 @@
 # When adding fields to a class in an __init__ method, we need to ensure that
 # field is copied in the copy() method.
 
+from __future__ import print_function
 
 import ast
 import collections
 import linecache
 from cPickle import loads, dumps
 import zlib
+import weakref
 
 import renpy.display
 import renpy.pyanalysis
@@ -43,13 +45,15 @@ from renpy.display.predict import displayable as predict_displayable
 
 from renpy.python import py_eval_bytecode
 from renpy.pyanalysis import Analysis, NOT_CONST, GLOBAL_CONST, ccache
+
 import hashlib
+import time
 
 # This file contains the abstract syntax tree for a screen language
 # screen.
 
 # A serial number that makes each SLNode unique.
-serial = 0
+serial = int(time.time() * 1000000)
 
 # A sentinel used to indicate we should use the value found in the
 # expression.
@@ -101,7 +105,15 @@ class SLContext(renpy.ui.Addable):
 
         # A cache associated with this context. The cache maps from
         # statement serial to information associated with the statement.
-        self.cache = { }
+        self.new_cache = { }
+
+        # The old cache, used to take information from the old version of
+        # this displayable.
+        self.old_cache = { }
+
+        # The miss cache, used to take information that isn't present in
+        # old_cache.
+        self.miss_cache = { }
 
         # The number of times a particular use statement has been called
         # in the current screen. We use this to generate a unique name for
@@ -148,6 +160,10 @@ class SLContext(renpy.ui.Addable):
         # to speed things up.
         self.unlikely = False
 
+        # The old and new generations of the use_cache.
+        self.new_use_cache = { }
+        self.old_use_cache = { }
+
     def add(self, d, key):
         self.children.append(d)
 
@@ -285,6 +301,7 @@ class SLNode(object):
 
         return False
 
+
 # A sentinel used to indicate a keyword argument was not given.
 NotGiven = renpy.object.Sentinel("NotGiven")
 
@@ -476,6 +493,7 @@ class SLCache(object):
         # The style prefix used when this statement was first created.
         self.style_prefix = None
 
+
 # A magic value that, if returned by a displayable function, is not added to
 # the parent.
 NO_DISPLAYABLE = renpy.display.layout.Null()
@@ -646,10 +664,12 @@ class SLDisplayable(SLBlock):
 
         screen = renpy.ui.screen
 
-        cache = context.cache.get(self.serial, None)
+        cache = context.old_cache.get(self.serial, None) or context.miss_cache.get(self.serial, None)
 
-        if cache is None:
-            context.cache[self.serial] = cache = SLCache()
+        if not isinstance(cache, SLCache):
+            cache = SLCache()
+
+        context.new_cache[self.serial] = cache
 
         copy_on_change = cache.copy_on_change
 
@@ -805,7 +825,7 @@ class SLDisplayable(SLBlock):
                 if self.scope:
                     keywords["scope"] = ctx.scope
 
-                if self.replaces and context.updating:
+                if self.replaces and ctx.updating:
                     keywords['replaces'] = old_main
 
                 # Pass the context
@@ -858,6 +878,8 @@ class SLDisplayable(SLBlock):
 
         finally:
 
+            ctx.keywords = None
+
             stack.pop()
 
             if imagemap:
@@ -1084,7 +1106,7 @@ class SLDisplayable(SLBlock):
     def copy_on_change(self, cache):
         c = cache.get(self.serial, None)
 
-        if c is not None:
+        if isinstance(c, SLCache):
             c.copy_on_change = True
 
         for i in self.children:
@@ -1176,7 +1198,6 @@ class SLIf(SLNode):
             context.predicted.add(self.serial)
 
         for cond, block in self.prepared_entries:
-
             try:
                 cond_value = (cond is None) or eval(cond, context.globals, context.scope)
             except:
@@ -1387,8 +1408,17 @@ class SLFor(SLBlock):
 
             value = [ 0 ]
 
-        newcaches = {}
-        oldcaches = context.cache.get(self.serial, newcaches)
+        newcaches = { }
+
+        oldcaches = context.old_cache.get(self.serial, newcaches) or { }
+
+        if not isinstance(oldcaches, dict):
+            oldcaches = { }
+
+        misscaches = context.miss_cache.get(self.serial, newcaches) or { }
+
+        if not isinstance(misscaches, dict):
+            misscaches = { }
 
         ctx = SLContext(context)
 
@@ -1396,13 +1426,17 @@ class SLFor(SLBlock):
 
             ctx.scope[variable] = v
 
-            cache = oldcaches.get(index, None)
+            ctx.old_cache = oldcaches.get(index, None) or { }
 
-            if cache is None:
-                cache = {}
+            if not isinstance(ctx.old_cache, dict):
+                ctx.old_cache = {}
 
-            newcaches[index] = cache
-            ctx.cache = cache
+            ctx.miss_cache = misscaches.get(index, None) or { }
+
+            if not isinstance(ctx.miss_cache, dict):
+                ctx.miss_cache = {}
+
+            newcaches[index] = ctx.new_cache = { }
 
             # Inline of SLBlock.execute.
 
@@ -1416,7 +1450,7 @@ class SLFor(SLBlock):
             if context.unlikely:
                 break
 
-        context.cache[self.serial] = newcaches
+        context.new_cache[self.serial] = newcaches
 
         if ctx.fail:
             context.fail = True
@@ -1426,7 +1460,8 @@ class SLFor(SLBlock):
 
     def copy_on_change(self, cache):
         c = cache.get(self.serial, None)
-        if c is None:
+
+        if not isinstance(c, dict):
             return
 
         for child_cache in c.values():
@@ -1575,7 +1610,12 @@ class SLUse(SLNode):
         target = renpy.display.screen.get_screen_variant(self.target)
 
         if target is None:
-            raise Exception("A screen named {} does not exist.".format(self.target))
+            self.constant = NOT_CONST
+
+            if renpy.config.developer:
+                raise Exception("A screen named {} does not exist.".format(self.target))
+            else:
+                return
 
         if target.ast is None:
             self.constant = NOT_CONST
@@ -1620,40 +1660,29 @@ class SLUse(SLNode):
 
         # Figure out the cache to use.
 
-        # True if we want to force-mark this as an update.
-        update = False
+        ctx = SLContext(context)
+        ctx.new_cache = context.new_cache[self.serial] = { }
+        ctx.miss_cache = context.miss_cache.get(self.serial, None) or { }
 
-        if (not context.predicting) and self.id:
+        if self.id:
 
-            # If we have an id, look it up in the current screen's use_cache.
-
-            current_screen = renpy.display.screen.current_screen()
             use_id = (self.target, eval(self.id, context.globals, context.scope))
 
-            cache = current_screen.use_cache.get(use_id, None)
+            ctx.old_cache = context.old_use_cache.get(use_id, None) or context.old_cache.get(self.serial, None) or { }
 
-            if cache is not None:
-                update = True
+            if use_id in ctx.old_use_cache:
+                ctx.updating = True
 
-            else:
-
-                if cache is None:
-                    cache = context.cache.get(self.serial, None)
-
-                if cache is None:
-                    cache = { }
-
-            context.cache[self.serial] = cache
-            current_screen.use_cache[use_id] = cache
+            ctx.new_use_cache[use_id] = ctx.new_cache
 
         else:
 
-            # Otherwise, look up the cache based on the statement's location.
+            ctx.old_cache = context.old_cache.get(self.serial, None) or { }
 
-            cache = context.cache.get(self.serial, None)
-
-            if cache is None:
-                context.cache[self.serial] = cache = { }
+        if not isinstance(ctx.old_cache, dict):
+            ctx.old_cache = { }
+        if not isinstance(ctx.miss_cache, dict):
+            ctx.miss_cache = { }
 
         # Evaluate the arguments.
         try:
@@ -1673,12 +1702,8 @@ class SLUse(SLNode):
         if ast.parameters is not None:
             new_scope = ast.parameters.apply(args, kwargs, ignore_errors=context.predicting)
 
-            scope = cache.get("scope", None)
-
-            if scope is None:
-                scope = cache["scope"] = new_scope
-            else:
-                scope.update(new_scope)
+            scope = ctx.old_cache.get("scope", None) or ctx.miss_cache.get("scope", None) or { }
+            scope.update(new_scope)
 
         else:
 
@@ -1691,17 +1716,15 @@ class SLUse(SLNode):
         scope["_scope"] = scope
 
         # Run the child screen.
-        ctx = SLContext(context)
         ctx.scope = scope
-        ctx.cache = cache
-        ctx.parent = context
-
-        if update:
-            ctx.updating = True
+        ctx.parent = weakref.ref(context)
 
         ctx.transclude = self.block
 
-        ast.execute(ctx)
+        try:
+            ast.execute(ctx)
+        finally:
+            del scope["_scope"]
 
         if ctx.fail:
             context.fail = True
@@ -1734,16 +1757,21 @@ class SLTransclude(SLNode):
         if not context.transclude:
             return
 
-        cache = context.cache.get(self.serial, None)
+        parent = context.parent
+        if parent is not None:
+            parent = parent()
 
-        if cache is None:
-            context.cache[self.serial] = cache = { }
+        ctx = SLContext(parent)
+        ctx.new_cache = context.new_cache[self.serial] = { }
+        ctx.old_cache = context.old_cache.get(self.serial, None) or { }
+        ctx.miss_cache = context.miss_cache.get(self.serial, None) or { }
 
-        cache["transclude"] = context.transclude
+        if not isinstance(ctx.old_cache, dict):
+            ctx.old_cache = { }
+        if not isinstance(ctx.miss_cache, dict):
+            ctx.miss_cache = { }
 
-        ctx = SLContext(context.parent)
-
-        ctx.cache = cache
+        ctx.new_cache["transclude"] = context.transclude
 
         ctx.children = context.children
         ctx.showif = context.showif
@@ -1971,6 +1999,7 @@ class SLScreen(SLBlock):
             debug = True
 
         context = SLContext()
+
         context.scope = scope
         context.globals = renpy.python.store_dicts["store"]
         context.debug = debug
@@ -1979,20 +2008,30 @@ class SLScreen(SLBlock):
 
         name = scope["_name"]
 
-        main_cache = current_screen.cache
+        def get_cache(d):
+            rv = d.get(name, None)
 
-        cache = main_cache.get(name, None)
-        if cache is None or (cache["version"] != self.version):
-            cache = { "version" : self.version }
-            main_cache[name] = cache
+            if (not isinstance(rv, dict)) or (rv.get("version", None) != self.version):
+                rv = { "version" : self.version }
+                d[name] = rv
 
-        context.cache = cache
+            return rv
+
+        context.old_cache = get_cache(current_screen.cache)
+        context.miss_cache = get_cache(current_screen.miss_cache)
+        context.new_cache = { "version" : self.version }
+
+        context.old_use_cache = current_screen.use_cache
+        context.new_use_cache = { }
 
         self.const_ast.execute(context)
 
         for i in context.children:
             renpy.ui.implicit_add(i)
 
+        current_screen.cache[name] = context.new_cache
+        current_screen.use_cache = context.new_use_cache
+
 
 class ScreenCache(object):
 
@@ -2004,6 +2043,7 @@ class ScreenCache(object):
 
         self.updated = False
 
+
 scache = ScreenCache()
 
 CACHE_FILENAME = "cache/screens.rpyb"
