@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,10 +19,11 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+import collections
 import renpy.sl2
 import renpy.sl2.slast as slast
 
-# A list of style prefixes that we know of.
+# A tuple of style prefixes that we know of.
 STYLE_PREFIXES = [
     '',
     'insensitive_',
@@ -48,6 +49,7 @@ all_statements = [ ]
 # Statements that can contain children.
 childbearing_statements = set()
 
+
 class Positional(object):
     """
     This represents a positional parameter to a function.
@@ -59,8 +61,10 @@ class Positional(object):
         if parser:
             parser.add(self)
 
-# Used to generate the documentation
-all_keyword_names = set()
+
+# This is a map from (prefix, use_style_prefixes) to a set of property names.
+properties = collections.defaultdict(set)
+
 
 class Keyword(object):
     """
@@ -70,11 +74,12 @@ class Keyword(object):
     def __init__(self, name):
         self.name = name
 
-        all_keyword_names.add(self.name)
+        properties['', False].add(name)
 
         if parser:
             parser.add(self)
 
+
 class Style(object):
     """
     This represents a style parameter to a function.
@@ -83,8 +88,7 @@ class Style(object):
     def __init__(self, name):
         self.name = name
 
-        for j in STYLE_PREFIXES:
-            all_keyword_names.add(j + self.name)
+        properties['', True].add(self.name)
 
         if parser:
             parser.add(self)
@@ -99,8 +103,7 @@ class PrefixStyle(object):
         self.prefix = prefix
         self.name = name
 
-        for j in STYLE_PREFIXES:
-            all_keyword_names.add(prefix + j + self.name)
+        properties[prefix, True].add(self.name)
 
         if parser:
             parser.add(self)
@@ -108,6 +111,11 @@ class PrefixStyle(object):
 
 class Parser(object):
 
+    # The number of children this statement takes, out of 0, 1, or "many".
+    # This defaults to "many" so the has statement errors out when not
+    # inside something that takes a single child.
+    nchildren = "many"
+
     def __init__(self, name):
 
         # The name of this object.
@@ -205,7 +213,7 @@ class Parser(object):
             screens.
 
         `block_only`
-            If true, only parse the
+            If true, only parse the block and not the initial properties.
         """
 
         seen_keywords = set()
@@ -233,7 +241,6 @@ class Parser(object):
 
             seen_keywords.add(name)
 
-
             expr = l.comma_expression()
 
             target.keyword.append((name, expr))
@@ -261,7 +268,6 @@ class Parser(object):
 
                 parse_keyword(l, 'expected a keyword argument, colon, or end of line.')
 
-
         # The index of the child we're adding to this statement.
         child_index = 0
 
@@ -285,8 +291,8 @@ class Parser(object):
                 loc = l.get_location()
 
                 if l.keyword(r'has'):
-                    if self.nchildren != 1:
-                        l.error("The %s statement does not take a layout." % self.name)
+                    if not can_has:
+                        l.error("The has statement is not allowed here.")
 
                     if child_index != 0:
                         l.error("The has statement may not be given after a child has been supplied.")
@@ -320,7 +326,6 @@ class Parser(object):
                 while not l.eol():
                     parse_keyword(l, "expected a keyword argument or end of line.")
 
-
     def add_positional(self, name):
         global parser
         parser = self
@@ -368,6 +373,7 @@ class Parser(object):
 def add(thing):
     parser.add(thing)
 
+
 # A singleton value.
 many = renpy.object.Sentinel("many")
 
@@ -483,9 +489,8 @@ def register_sl_displayable(*args, **kwargs):
 class DisplayableParser(Parser):
 
     def __init__(self, name, displayable, style, nchildren=0, scope=False,
-        pass_context=False, imagemap=False, replaces=False, default_keywords={},
-        hotspot=False, default_properties=True):
-
+                 pass_context=False, imagemap=False, replaces=False, default_keywords={},
+                 hotspot=False, default_properties=True):
         """
         `scope`
             If true, the scope is passed into the displayable functionas a keyword
@@ -536,7 +541,6 @@ class DisplayableParser(Parser):
             add(renpy.sl2.slproperties.ui_properties)
             add(renpy.sl2.slproperties.position_properties)
 
-
     def parse_layout(self, loc, l, parent):
         return self.parse(loc, l, parent, True)
 
@@ -575,6 +579,7 @@ class DisplayableParser(Parser):
 
         return rv
 
+
 class IfParser(Parser):
 
     def __init__(self, name, node_type, parent_contents):
@@ -595,7 +600,6 @@ class IfParser(Parser):
         if not parent_contents:
             childbearing_statements.add(self)
 
-
     def parse(self, loc, l, parent):
 
         if self.parent_contents:
@@ -652,6 +656,7 @@ class IfParser(Parser):
 
         return rv
 
+
 if_statement = IfParser("if", slast.SLIf, True)
 IfParser("showif", slast.SLShowIf, False)
 
@@ -676,6 +681,7 @@ class ForParser(Parser):
             if l.match(r"\("):
                 name = self.name_or_tuple_pattern(l)
                 l.require(r'\)')
+                pattern = True
             else:
                 name = l.name()
 
@@ -726,6 +732,7 @@ class ForParser(Parser):
 
         return rv
 
+
 ForParser("for")
 
 
@@ -734,7 +741,7 @@ class OneLinePythonParser(Parser):
     def parse(self, loc, l, parent):
 
         loc = l.get_location()
-        source = l.require(l.rest)
+        source = l.require(l.rest_statement)
 
         l.expect_eol()
         l.expect_noblock("one-line python")
@@ -742,6 +749,7 @@ class OneLinePythonParser(Parser):
         code = renpy.ast.PyCode(source, loc)
         return slast.SLPython(loc, code)
 
+
 OneLinePythonParser("$")
 
 
@@ -761,6 +769,7 @@ class MultiLinePythonParser(Parser):
         code = renpy.ast.PyCode(source, loc)
         return slast.SLPython(loc, code)
 
+
 MultiLinePythonParser("python")
 
 
@@ -772,6 +781,7 @@ class PassParser(Parser):
 
         return slast.SLPass(loc)
 
+
 PassParser("pass")
 
 
@@ -788,6 +798,7 @@ class DefaultParser(Parser):
 
         return slast.SLDefault(loc, name, rest)
 
+
 DefaultParser("default")
 
 
@@ -812,7 +823,7 @@ class UseParser(Parser):
             l.expect_block("use statement")
 
             block = slast.SLBlock(loc)
-            self.parse_contents(l, block, block_only=True)
+            self.parse_contents(l, block, can_has=True, block_only=True)
 
         else:
             l.expect_eol()
@@ -822,16 +833,19 @@ class UseParser(Parser):
 
         return slast.SLUse(loc, target, args, id_expr, block)
 
+
 UseParser("use")
 Keyword("style_prefix")
 Keyword("style_group")
 
+
 class TranscludeParser(Parser):
 
     def parse(self, loc, l, parent):
         l.expect_eol()
         return slast.SLTransclude(loc)
 
+
 TranscludeParser("transclude")
 
 
@@ -944,6 +958,7 @@ class ScreenParser(Parser):
 
         return screen
 
+
 screen_parser = ScreenParser()
 Keyword("modal")
 Keyword("zorder")
@@ -952,6 +967,8 @@ Keyword("predict")
 Keyword("style_group")
 Keyword("style_prefix")
 Keyword("layer")
+parser = None
+
 
 def init():
     screen_parser.add(all_statements)
@@ -970,4 +987,3 @@ def parse_screen(l, loc):
     """
 
     return screen_parser.parse(loc, l, None)
-
